/*
 * ESP32 BMP/Sound/PCM file player for ILI9341
 *   BMP : Image data(*.bmp)
 *   SND : Sound data(*.snd)
 *       Hz/pitch time(ms)
 *   PCM : PCM sound(*.pcm)
 *       raw, unsigned 8bit, 11.025kHz
 */
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <Adafruit_ILI9341.h>
#include "pitches.h"

#define TFT_DC  2
#define TFT_CS 15
#define TFT_RST 4

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int SoundPin = 25;
static int tone_init = 0;

#define SD_CS 5
#define MAX_FILE 16
#define MAX_SND 1024
File root;
char fname0[MAX_FILE][32];
char fnum0;
char fname1[MAX_FILE][32];
char fnum1;
char fname2[MAX_FILE][32];
char fnum2;
int *snd_pitch;
int *snd_leng;

int  num, old_num;

int StickXPin = 34;
int StickYPin = 35;
int StickX;
int StickY;

int ButtonAPin = 32;
int ButtonBPin = 33;
int ButtonA;
int ButtonB;
int OldA;
int OldB;

// 0:Select  1:BMP select  2:SND select  3:BMP view  4:SND play
int ftype;

void setup(void) {
  Serial.begin(115200);

  tft.begin(40000000);
  tft.setRotation(0);
  _width=tft.width();
  _height=tft.height();
  
  yield();

  Serial.print("Initializing SD card...");
  pinMode(SD_CS, PULLUP);
  pinMode(MISO, PULLUP);
  if (!SD.begin(SD_CS)) {
    Serial.println("failed!");
    show_no_sd();
    for(;;);
  }
  root = SD.open("/");

  fnum0=0;
  fnum1=0;
  fnum2=0;
  printDirectory(root, 0, ".bmp", ".snd", ".pcm");

  Serial.println("OK!");

  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);

  snd_pitch=(int *)malloc(MAX_SND*sizeof(int));
  snd_leng=(int *)malloc(MAX_SND*sizeof(int));

  ftype=0;
  num=0;
  old_num=-1;
}

void loop() {
  int push_btn;

  StickX = analogRead(StickXPin)/4;
  StickY = analogRead(StickYPin)/4;
  OldA=ButtonA;
  OldB=ButtonB;
  ButtonA = digitalRead(ButtonAPin);
  ButtonB = digitalRead(ButtonBPin);

// Menu
  if (ftype==0){
    show_select();
    old_num=num;
    if (StickY<256){
      num++;
      delay(200);
    }
    if (StickY>768){
      num--;
      delay(200);
    }
    if (num<0)
      num=0;
    if (num>2)
      num=2;

    if (ButtonA==0 && OldA==1){
      ftype=num+1;
      num=0;
      old_num=-1;
    }

// BMP
  }else if (ftype==1){
    show_bmp_sel();
    old_num=num;
    if (StickY<256){
      num++;
      delay(200);
    }
    if (StickY>768){
      num--;
      delay(200);
    }
    if (num<0)
      num=0;
    if (num>=fnum0)
      num=fnum0-1;

    if (ButtonA==0 && OldA==1){
      ftype=11;
    }

    if (ButtonB==0 && OldB==1){
      ftype=0;
      num=0;
      old_num=-1;
    }
  }else if (ftype==11){
    tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
    bmpDraw(fname0[num], 0, 0);
    ftype=21;
  }else if (ftype==21){
    if (ButtonA==0 && OldA==1){
      ftype=1;
      old_num=-1;
    }
    if (ButtonB==0 && OldB==1){
      ftype=1;
      old_num=-1;
    }

// SND
  }else if (ftype==2){
    show_snd_sel();
    old_num=num;
    if (StickY<256){
      num++;
      delay(200);
    }
    if (StickY>768){
      num--;
      delay(200);
    }
    if (num<0)
      num=0;
    if (num>=fnum1)
      num=fnum1-1;

    if (ButtonA==0 && OldA==1){
      ftype=12;
    }

    if (ButtonB==0 && OldB==1){
      ftype=0;
      num=0;
      old_num=-1;
    }
  }else if (ftype==12){
    tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
    SoundPlay(fname1[num]);
    ftype=2;
    old_num=-1;

// PCM
  }else if (ftype==3){
    show_pcm_sel();
    old_num=num;
    if (StickY<256){
      num++;
      delay(200);
    }
    if (StickY>768){
      num--;
      delay(200);
    }
    if (num<0)
      num=0;
    if (num>=fnum2)
      num=fnum2-1;

    if (ButtonA==0 && OldA==1){
      ftype=13;
    }

    if (ButtonB==0 && OldB==1){
      ftype=0;
      num=0;
      old_num=-1;
    }
  }else if (ftype==13){
    tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
    PcmPlay(fname2[num]);
    ftype=3;
    old_num=-1;

  }else{
    ftype=0;
  }
  delay(50);
  noTone(SoundPin);
}

void printDirectory(File dir, int numTabs,
                    char *typestr0, char *typestr1, char *typestr2 ) {
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (fnum0<MAX_FILE && stristr(entry.name(), typestr0)!=NULL){
      strcpy(fname0[fnum0++], entry.name());
    }else if (fnum1<MAX_FILE && stristr(entry.name(), typestr1)!=NULL){
      strcpy(fname1[fnum1++], entry.name());
    }else if (fnum2<MAX_FILE && stristr(entry.name(), typestr2)!=NULL){
      strcpy(fname2[fnum2++], entry.name());
    }else if(fnum0>=MAX_FILE && fnum1>=MAX_FILE && fnum2>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr0, typestr1, typestr2);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

// This function opens a Windows Bitmap (BMP) file and
// displays it at the given coordinates.  It's sped up
// by reading many pixels worth of data at a time
// (rather than pixel by pixel).  Increasing the buffer
// size takes more of the Arduino's precious RAM but
// makes loading a little faster.  20 pixels seems a
// good balance.

#define BUFFPIXEL 20

void bmpDraw(char *filename, int16_t x, int16_t y) {

  File     bmpFile;
  int      bmpWidth, bmpHeight;   // W+H in pixels
  uint8_t  bmpDepth;              // Bit depth (currently must be 24)
  uint32_t bmpImageoffset;        // Start of image data in file
  uint32_t rowSize;               // Not always = bmpWidth; may have padding
  uint8_t  sdbuffer[3*BUFFPIXEL]; // pixel buffer (R+G+B per pixel)
  uint8_t  buffidx = sizeof(sdbuffer); // Current position in sdbuffer
  boolean  goodBmp = false;       // Set to true on valid header parse
  boolean  flip    = true;        // BMP is stored bottom-to-top
  int      w, h, row, col, x2, y2, bx1, by1;
  uint8_t  r, g, b;
  uint32_t pos = 0, startTime = millis();

  if((x >= tft.width()) || (y >= tft.height())) return;

  Serial.println();
  Serial.print(F("Loading image '"));
  Serial.print(filename);
  Serial.println('\'');

  // Open requested file on SD card
  if ((bmpFile = SD.open(filename)) == NULL) {
    Serial.print(F("File not found"));
    return;
  }

  // Parse BMP header
  if(read16(bmpFile) == 0x4D42) { // BMP signature
    Serial.print(F("File size: ")); Serial.println(read32(bmpFile));
    (void)read32(bmpFile); // Read & ignore creator bytes
    bmpImageoffset = read32(bmpFile); // Start of image data
    Serial.print(F("Image Offset: ")); Serial.println(bmpImageoffset, DEC);
    // Read DIB header
    Serial.print(F("Header size: ")); Serial.println(read32(bmpFile));
    bmpWidth  = read32(bmpFile);
    bmpHeight = read32(bmpFile);
    if(read16(bmpFile) == 1) { // # planes -- must be '1'
      bmpDepth = read16(bmpFile); // bits per pixel
      Serial.print(F("Bit Depth: ")); Serial.println(bmpDepth);
      if((bmpDepth == 24) && (read32(bmpFile) == 0)) { // 0 = uncompressed

        goodBmp = true; // Supported BMP format -- proceed!
        Serial.print(F("Image size: "));
        Serial.print(bmpWidth);
        Serial.print('x');
        Serial.println(bmpHeight);

        // BMP rows are padded (if needed) to 4-byte boundary
        rowSize = (bmpWidth * 3 + 3) & ~3;

        // If bmpHeight is negative, image is in top-down order.
        // This is not canon but has been observed in the wild.
        if(bmpHeight < 0) {
          bmpHeight = -bmpHeight;
          flip      = false;
        }

        // Crop area to be loaded
        x2 = x + bmpWidth  - 1; // Lower-right corner
        y2 = y + bmpHeight - 1;
        if((x2 >= 0) && (y2 >= 0)) { // On screen?
          w = bmpWidth; // Width/height of section to load/display
          h = bmpHeight;
          bx1 = by1 = 0; // UL coordinate in BMP file
          if(x < 0) { // Clip left
            bx1 = -x;
            x   = 0;
            w   = x2 + 1;
          }
          if(y < 0) { // Clip top
            by1 = -y;
            y   = 0;
            h   = y2 + 1;
          }
          if(x2 >= tft.width())  w = tft.width()  - x; // Clip right
          if(y2 >= tft.height()) h = tft.height() - y; // Clip bottom
  
          // Set TFT address window to clipped image bounds
          tft.startWrite(); // Requires start/end transaction now
//          tft.setAddrWindow(x, y, w, h);
  
          for (row=0; row<h; row++) { // For each scanline...
  
            // Seek to start of scan line.  It might seem labor-
            // intensive to be doing this on every line, but this
            // method covers a lot of gritty details like cropping
            // and scanline padding.  Also, the seek only takes
            // place if the file position actually needs to change
            // (avoids a lot of cluster math in SD library).
            if(flip) // Bitmap is stored bottom-to-top order (normal BMP)
              pos = bmpImageoffset + (bmpHeight - 1 - (row + by1)) * rowSize;
            else     // Bitmap is stored top-to-bottom
              pos = bmpImageoffset + (row + by1) * rowSize;
            pos += bx1 * 3; // Factor in starting column (bx1)
            if(bmpFile.position() != pos) { // Need seek?
              tft.endWrite(); // End TFT transaction
              bmpFile.seek(pos);
              buffidx = sizeof(sdbuffer); // Force buffer reload
              tft.startWrite(); // Start new TFT transaction
            }
            for (col=0; col<w; col++) { // For each pixel...
              // Time to read more pixel data?
              if (buffidx >= sizeof(sdbuffer)) { // Indeed
                tft.endWrite(); // End TFT transaction
                bmpFile.read(sdbuffer, sizeof(sdbuffer));
                buffidx = 0; // Set index to beginning
                tft.startWrite(); // Start new TFT transaction
              }
              // Convert pixel from BMP to TFT format, push to display
              b = sdbuffer[buffidx++];
              g = sdbuffer[buffidx++];
              r = sdbuffer[buffidx++];
//              tft.writePixel(tft.color565(r,g,b));
              tft.writePixel(col, row, tft.color565(r,g,b));
            } // end pixel
          } // end scanline
          tft.endWrite(); // End last TFT transaction
        } // end onscreen
        Serial.print(F("Loaded in "));
        Serial.print(millis() - startTime);
        Serial.println(" ms");
      } // end goodBmp
    }
  }

  bmpFile.close();
  if(!goodBmp) Serial.println(F("BMP format not recognized."));
}

// These read 16- and 32-bit types from the SD card file.
// BMP data is stored little-endian, Arduino is little-endian too.
// May need to reverse subscript order if porting elsewhere.

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

void SoundPlay(char *filename)
{
  int i;
  uint8_t fbuf;
  int snd_num;
  int num, num_sign, n, ptch;
  File     sndFile;

  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 0);
  tft.print(filename);

  if ((sndFile = SD.open(filename)) == NULL) {
    Serial.print(F("File not found"));
    return;
  }

  snd_num=0;
  i=0;
  num=0;
  num_sign=1;
  while (sndFile.available()) {
    fbuf=sndFile.read();
    if ('0'<=fbuf && fbuf<='9'){
      num=num*10+fbuf-'0';
    }else if (fbuf=='-'){
      num_sign=-1;
    }else if ('A'<=fbuf && fbuf<='G'){
      num=1000+(fbuf-'A')*100;
    }else if (fbuf=='S'){
      num=num+10;
    }else if (num>0){
      if (i==0){
        if (num>=10000){
          n=(num/100) % 100;
          switch(n){
            case 20 : ptch=8372;  break; // C9
            case 21 : ptch=8870;  break; // CS9
            case 30 : ptch=9397;  break; // D9
            case 31 : ptch=9956;  break; // DS9
            case 40 : ptch=10548; break; // E9
            case 50 : ptch=11175; break; // F9
            case 51 : ptch=11840; break; // FS9
            case 60 : ptch=12544; break; // G9
            case 61 : ptch=13290; break; // GS9
            case 00 : ptch=14080; break; // A9
            case 01 : ptch=14917; break; // AS9
            case 10 : ptch=15804; break; // B9
            default : ptch=-10000; break;
          }
          n=num % 10;
          n=1 << (9-n);
          ptch=(ptch+n/2)/n;
          snd_pitch[snd_num]=ptch;
        }else
          snd_pitch[snd_num]=num*num_sign;
        i=1;
      }else{
        if (num>=10000 || num<10)
          snd_leng[snd_num]=10;
        else
          snd_leng[snd_num]=num*num_sign;
        Serial.print(snd_num);
        Serial.print(" ");
        Serial.print(snd_pitch[snd_num]);
        Serial.print(" ");
        Serial.println(snd_leng[snd_num]);
        snd_num++;
        i=0;
      }
      num=0;
      num_sign=1;
    }
  }
  sndFile.close();

  for(i=0;i<snd_num;i++){
    if (snd_pitch[i]>0)
      tone(SoundPin, snd_pitch[i], snd_leng[i]);
    if (20<snd_leng[i] && snd_leng[i]<10000)
      delay(snd_leng[i]-20);
    else
      delay(20);
    noTone(SoundPin);
    delay(20);

  }
}

void PcmPlay(char *filename)
{
  int i;
  int pcm_num;
  int val;
  uint8_t *pcm_data0;
  uint8_t *pcm_data1;
  uint8_t *fbuf;
  File     pcmFile;

  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 0);
  tft.print(filename);

  if ((pcmFile = SD.open(filename)) == NULL) {
    Serial.print(F("File not found"));
    return;
  }

  while (pcmFile.available()) {
    val=pcmFile.read();
// Volume x1.25
//    val=(val-128)*20/16+128;
// Volume x1.5
    val=(val-128)*24/16+128;

    if (val<0) val=0;
    if (val>255) val=255;

    dacWrite(SoundPin, val);
    delayMicroseconds(91-27); // 11.025kHz
//    delayMicroseconds(46-27); // 22.05kHz
  }
  pcmFile.close();
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void show_no_sd()
{
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_RED);
  tft.setCursor(16, 0);
  tft.print("Error");

  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 32);
  tft.print("No SD card");
}

void show_select()
{
  if (num==old_num)
    return;
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(16, 0);
  tft.print("Select");

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.fillRect(0, 32+num*16, _width, 8, ILI9341_RED);
  tft.setCursor(0, 32);
  tft.print("1.Image file(*.BMP)");
  tft.setCursor(0, 48);
  tft.print("2.Sound file(*.SND)");  
  tft.setCursor(0, 64);
  tft.print("3.PCM file(*.PCM)");  
}

void show_bmp_sel()
{
  int i;
  
  if (num==old_num)
    return;
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(8, 0);
  tft.print("BMP file");

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  if (fnum0<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+num*8, _width, 8, ILI9341_RED);
  for(i=0;i<fnum0;i++){
    tft.setCursor(0, 16+i*8);
    tft.print(fname0[i]);
  }
}

void show_snd_sel()
{
  int i;
  
  if (num==old_num)
    return;
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(8, 0);
  tft.print("SND file");

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  if (fnum1<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+num*8, _width, 8, ILI9341_RED);
  for(i=0;i<fnum1;i++){
    tft.setCursor(0, 16+i*8);
    tft.print(fname1[i]);
  }
}

void show_pcm_sel()
{
  int i;
  
  if (num==old_num)
    return;
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(8, 0);
  tft.print("PCM file");

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  if (fnum2<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+num*8, _width, 8, ILI9341_RED);
  for(i=0;i<fnum2;i++){
    tft.setCursor(0, 16+i*8);
    tft.print(fname2[i]);
  }
}

void tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  if (tone_init==0){
    ledcSetup(1, 12000, 8);
    ledcAttachPin(pin, 1);
    tone_init=1;
  }
  ledcWriteTone(1, freq);
}

void noTone (uint8_t pin)
{
  ledcWriteTone(1, 1);
}
